
/*
 *    cameh.c   << Device Driver for DSP 6002>>    --10 Jul. 1998--
 *    (based on camdrv - another Device Driver for DSP6002)
 *    the present driver includes an emulator to the 
 *    Event Handler instructions
 *    Author   : R.V. Ribas
 *    Version for kernel 2.6.x Dez. 2005
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/ioctl.h>
#include <linux/ioport.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/delay.h>
#include "camdrv-eh.h"


MODULE_LICENSE("GPL");
MODULE_AUTHOR("R. V. Ribas");
MODULE_DESCRIPTION("Driver for CC ISA DSP6002");

#define BUFFER_SIZE 16384
#define INSTR_SIZE  4096
#define ENABLE 0x0c03
#define DISABLE 0x0d03
#define ENABLE_CAMEH 0
#define DISABLE_CAMEH 4

static int iev=0,MAXSIZE=512,FLAG[16],NEXTW,NEXTR,NBUF,nwords=0;
static int Running=0,lost=0,lock=0;
static unsigned int offset_b=0;
static volatile unsigned short int cc_buf[BUFFER_SIZE];
static volatile unsigned int  cc_inst_buf[INSTR_SIZE]; 

static int cc_open(struct inode* inode, struct file* filep);
static ssize_t cc_read (struct file *, char *, size_t,loff_t *);
static ssize_t cc_write (struct file *, const char *, size_t, loff_t *);
static int cc_release(struct inode* inode, struct file* filep);
static int cc_ioctl(struct inode *inode, struct file *filep, unsigned int cmd, unsigned long arg);
irqreturn_t cc_interrupt(int irq, void* dev_id, struct pt_regs* regs);

static void set_crate_number(int);
static void camcl(int );
static int camo24 (int ,int ,int , int);
static int cami16 (int ,int ,int , int *);
static int cami24 (int ,int ,int , int *);
static int camo16 (int ,int ,int , int);
static void ev_read (unsigned short *);
static int NN,AA,FF,CC_IO_OP,NFA;
static int major = AC_MAJOR;

static struct file_operations cc_fops = {
    .owner = THIS_MODULE,
    .open = cc_open,
    .write = cc_write,
    .read = cc_read,
    .release = cc_release,
    .ioctl = cc_ioctl,
};

int cc_init(void);
void cc_exit(void);

module_init(cc_init);
module_exit(cc_exit);

static int cc_in_use[4];

int __init cc_init(void)
{
  int i;
  if (register_chrdev(major, CC_NAME, &cc_fops))
  {
    printk(" unable to get major %i for " CC_NAME " device\n",major);
    return -EIO;
  }
  for(i=0;i<4;i++) cc_in_use[i] = 0;
/*  release_region(CAMAC,16);
  free_irq(CC_IRQ,NULL);
*/
  printk (CC_NAME ": initializing module \n");
  return 0;

}

void __exit cc_exit(void)
{
   unregister_chrdev(major, CC_NAME);
   free_irq(CC_IRQ,NULL);
   release_region(CAMAC,16);
   printk(CC_NAME ": module removed\n");
   return;
}      

static int cc_open(struct inode* inode, struct file* filep)
{
 unsigned int crate = MINOR(inode->i_rdev);
 int istat;
  printk("Opening Crate # %d \n",crate);
  if (crate > 3){
    printk("Crate # must be <4\n");
    return -EBUSY;
    }
  if(cc_in_use[crate]){
    printk("Crate in Use...\n");
    return -EBUSY;
    }
  printk(" requesting ports...\n");
  if(!request_region(CAMAC,16,"CAMDRV")){
   printk(CC_NAME": Unable to register CAMAC ports\n");
   return -EIO; 
   }
printk(" requesting irq...\n");
  istat = request_irq (CC_IRQ, cc_interrupt,IRQF_DISABLED, "CAMDRV" ,NULL);
  if (istat){
    printk(CC_NAME": can't allocate IRQ%d for DSP camac\n",CC_IRQ);
    release_region(CAMAC,16);
    return -EIO;
  }
  set_crate_number(crate);
  camcl(64);
  camcl(4);
  printk("Crate open OK  \n");
  cc_in_use[crate] = 1;
  return 0; 
}

static int cc_release(struct inode* inode, struct file* filep)
{
 unsigned int crate = MINOR(inode->i_rdev);
 cc_in_use[crate] = 0;
 release_region(CAMAC,16);
 free_irq(CC_IRQ,NULL);
 camcl(1);
 return 0;
}

static int cc_ioctl(struct inode* inode, struct file *filep, unsigned int cmd, unsigned long arg)
{
  int crate = MINOR(inode->i_rdev);
  unsigned short int dummy[64];
  int i,lam;

  if (crate > 3)
    return -ENODEV;

/* cmd validos: CCREAD16, CCREAD24, CCWRITE16, CCWRITE24, CCRESET
                CCREADEVNT CCWAITLAM CCMAXSIZE
*/
  switch (cmd) {
  case CCMAXSIZE:
    MAXSIZE=arg;
    if(MAXSIZE<256)MAXSIZE=256;
    if(MAXSIZE>4096)MAXSIZE=4096;
    NBUF=BUFFER_SIZE/MAXSIZE;
    if(NBUF > 16) NBUF=16;
    printk(" maxsize,nbuf = %d %d\n",MAXSIZE,NBUF);
    return 0;
    break;
  case CCRESET:
    printk(" CCRESET %d \n",(int) arg);
    if(arg == ENABLE || arg == ENABLE_CAMEH){
      lost=0;
      printk(" Running...\n");
      for(i=0;i<16;i++)FLAG[i]=0;
      nwords=0;
      offset_b=0;
      for(i=0;i<BUFFER_SIZE;i++)cc_buf[i]=0;
      NEXTW=0;
      NEXTR=0;
      Running = 1;
      lam=(inb_p(CAM8) & 0x007c);
      if(lam<=0x007b)ev_read(dummy);
      set_crate_number(crate);
      camcl ((int) ENABLE_CAMEH);
    }
    if(arg==DISABLE || arg == DISABLE_CAMEH)
    {
      printk(" Stopped...lost %d events\n",lost);
      set_crate_number(crate);
      camcl ((int) DISABLE_CAMEH);
    }
    
    return 0;
    break;
  case CCWAITLAM:
    printk("Invalid call: CCWAITLAM \n");
    return -1;
    break;
  case CCGETLOST:
    return lost;
    break;
  case CCGETUSED:
    return 0;
    break;
  case CCGETBUF:
    return 0;
    break;
  case CCSETCRATE:
    if(arg<4) set_crate_number(arg);
    return 0;
  default:
    NFA=arg;
    NN=arg&0x001f;
    FF=(arg&0x03e0)>>5;
    AA=(arg&0x3c00)>>10;
    CC_IO_OP=cmd;
    return (0);
  }
 
}

irqreturn_t cc_interrupt(int irq, void* dev, struct pt_regs* regs)
 {
  static unsigned short ev_buf[128];
  int i;
  iev=0;
  lock=0;
  ev_read(ev_buf);
  nwords=nwords+iev;
  if(nwords > (MAXSIZE-64)){
    FLAG[NEXTW++]=nwords;
    if(NEXTW >= NBUF) NEXTW=0;
    nwords=0;
    offset_b=NEXTW*MAXSIZE; 
  }
  if(FLAG[NEXTW]==0){
    for(i=0;i<iev;i++)cc_buf[offset_b+i]=ev_buf[i];
    offset_b=offset_b+iev;
  }else lost++;
  
  local_irq_enable();
  return IRQ_HANDLED;
}

static  ssize_t cc_read (struct file *file, char *buffer, size_t size, loff_t *loff)
{
  unsigned short int dummy[64];
  int result;
  int max,nbytes,QX;
  unsigned int Dat;
  unsigned short int *offset;

  /*  Check that we can really write to the buffer   */
  max=size;
  max=max>>1;
  if(max > BUFFER_SIZE){ 
   printk(CC_NAME": transfer block larger than 8K %d\n",size);
   return -1;
  }
  if (!(result = access_ok(VERIFY_WRITE,buffer,size)))
     return result;

  switch (CC_IO_OP){
  case CCREAD16:
    QX = cami16 (NN,AA,FF,&Dat);
    copy_to_user(buffer, &Dat, 2);
    return QX;
    break;
  case CCREAD24:
    QX = cami24 (NN,AA,FF,&Dat);
    copy_to_user(buffer, &Dat, 4);
    return QX;
    break;
  case CCREADEVNT:
    nbytes = FLAG[NEXTR]*2;
    if(nbytes > 0){
      offset = (short int *)cc_buf;
      FLAG[NEXTR]=0;
      offset=offset+MAXSIZE*NEXTR++;
      if(NEXTR >= NBUF) NEXTR = 0;
      copy_to_user(buffer, (void *)offset,nbytes);
      return nbytes;
    }else {
      lock++;
      if(lock > 20) {
	ev_read(dummy);
	lock=0;
	}
      return 0;
    }
    break;

    default:
    return -1;
  }
}

static ssize_t
cc_write (struct file *file, const char *buffer, size_t size, loff_t *loff)
{
  int result, Dat=0;
  unsigned short int dummy[64];

  /*
   * Check that we can really read from the buffer
   */
  if (!(result = access_ok(VERIFY_READ,buffer,size)))  return result;

  switch (CC_IO_OP){
  case CCWRITE16:
    /*    Dat=get_fs_long(buffer); */
    copy_from_user(&Dat,buffer,2);
    return camo16(NN,AA,FF,Dat);
    break;
  case CCWRITE24:
    /*    Dat=get_fs_long(buffer); */
    copy_from_user(&Dat,buffer,4);
    return camo24(NN,AA,FF,Dat);
    break;
  case CCLOADEH:
    copy_from_user((int *)cc_inst_buf,buffer,size);
    printk("loading %d instructions \n",size/4);
    camcl(4);
    iev=0;
    ev_read(dummy);
    return 0;
    break;
  default:
    return -1;
  }
}

static int camo16(int N,int A,int F,int D)
{
  outb_p((short) N,CAM5);
  outb_p((short) F,CAM4);
  outb_p((short) A,CAM3);
  outb_p((short) D,CAM2);
  outb_p((short) D>>8,CAM1);
  outb_p((short) N,CAM7);  /*  start camac cycle */
  return inb_p(CAM8);
}

static int camo24(int N,int A,int F,int D)
{
  
  outb_p((short) N,CAM5);
  outb_p((short) F,CAM4);
  outb_p((short) A,CAM3);
  outb_p((short) D,CAM2);
  outb_p((short) D>>8,CAM1);
  D=D>>16;
  outb_p((short) D,CAM);
  outb_p((short) N,CAM7);   /* start camac cycle */
  return inb_p(CAM8);
}
static int cami16(int N,int A,int F,int *D)
{
  int temp;
  outb_p((short) N,CAM5);
  outb_p((short) F,CAM4);
  outb_p((short) A,CAM3);
  outb_p((short) N,CAM7);  /* start camac cycle */
  temp=(inb_p(CAM11)&0x000000ff)+((inb_p(CAM10)<<8)&0x0000ff00);
  *D=(temp&0x0000ffff);
  return inb_p(CAM8);
}

static int cami24(int N,int A,int F,int *D)
{
  int temp;
  outb_p((short) N,CAM5);
  outb_p((short) F,CAM4);
  outb_p((short) A,CAM3);
  outb_p((short) N ,CAM7);   /* start camac cycle */ 

temp=(inb_p(CAM11)&0x000000ff)+((inb_p(CAM10)<<8)&0x0000ff00)+
                             ((inb_p(CAM9)<<16)&0x00ff0000);
  *D=temp&0x00ffffff;
  return inb_p(CAM8);
}

static void camcl(int I)
{

   outb_p((short) I,CAM6);
   if((I & 3) != 0) outb_p((short) I,CAM7);
   return;
}

static void set_crate_number(int I)
{
int mask=0;

  mask=mask&0x0000ffcf;
  mask=mask|(I<<4);
  outb_p((short) mask,CAM15);
}

static void ev_read (unsigned short int ev_buf[])
{
  int i,looping;
  unsigned short int par,PAT=0,EX=0,TXR=0;
  unsigned int CA=0;
  unsigned int inst,N,F,A,q;
  int hbit,il;
  looping=1;
  il=0;
  i=0;
  while ((looping == 1) && (i<4096) ){
    il++;
    if(il>4096){
       printk("EH programm Looping...\n");
       return;
    }
    if(iev > 127){
      printk("Large event truncated \n");
      return;
    }
    hbit=(cc_inst_buf[i] & 0x00f00000);
    if(hbit>0x200000){
      inst = (cc_inst_buf[i] & 0x007f0000);
      par  = cc_inst_buf[i++] & 0xffff;
    }
    else{
      inst = (cc_inst_buf[i] & 0x007fc000);
      par  = cc_inst_buf[i++] & 0x3fff;
    }
    if((inst == 0x100000)){
      
      N=par&0x001f;
      F=(par&0x03e0)>>5;
      A=(par&0x3c00)>>10;
      if((par&0x03e0) < 0x00e0){
        q=cami16(N,A,F,&CA);
      }
      else{
        q=camo16(N,A,F,CA);
      }
    }
    else if(inst == 0x200000)looping=0;               /* RTRN */
    else if(inst == 0x250000)i = par-1;               /* BRU A */
    else if(inst == 0x2A0000)looping=0;               /* INIT */
    else if(inst == 0x300000){if((PAT&par) != 0)i++;} /* SKIP PAT .ANY.#*/
    else if(inst == 0x320000){if((EX&par) != 0)i++;}  /* SKIP EX .ANY.#*/
    else if(inst == 0x330000){if((CA&par) != 0)i++;}  /* SKIP CA .ANY.#*/
    else if(inst == 0x340000){if((PAT&par) == 0)i++;} /* SKIP PAT .NONE.#*/
    else if(inst == 0x360000){if((EX&par) == 0)i++;}  /* SKIP EX .NONE.#*/
    else if(inst == 0x370000){if((CA&par) == 0)i++;}  /* SKIP CA .NONE.#*/
    else if(inst == 0x3b0000){if(CA<par) i++;}        /* SKIP CA .LT.# */
    else if(inst == 0x3f0000){if(CA>par) i++;}        /* SKIP CA .GT.# */
    else if(inst == 0x380000){if(PAT<par) i++;}       /* SKIP PAT .LT.#*/
    else if(inst == 0x3c0000){if(PAT>par) i++;}       /* SKIP PAT .GT.#*/
    else if(inst == 0x400000)CA = par;                /* MOV #,CA */
    else if(inst == 0x410000)PAT = CA;                /* MOV CA,PAT */
    else if(inst == 0x420000)CA = PAT;                /* MOV PAT,CA */
    else if(inst == 0x430000)EX = par;                /* SETX # */
    else if(inst == 0x440000)EX = EX +par;            /* ADDX # */
    else if(inst == 0x450000)CA = CA >> par;          /* SHFT # */
    else if(inst == 0x510000)TXR = CA;                /* MOV CA1,TXR */
    else if(inst == 0x580000)ev_buf[iev++] = par;     /* OUT # */
    else if(inst == 0x590000)ev_buf[iev++] = (CA&0x0fff); /* OUT CA */
    else if(inst == 0x5A0000)ev_buf[iev++] = PAT;     /* OUT PAT */
    else if(inst == 0x5C0000)ev_buf[iev++]=(TXR&par); /* MERG #,TXR */
    else if(inst == 0x600000)udelay(par&0x00ff);             /* DLAY # */
    else if(inst == 0x700000) par = 0;                /* CLRB */
    else if(inst == 0x710000) par = 0;                /* SETB */
    else if(inst == 0x000000) return;                 /* RETURN */
    else {
     printk(" Invalid instruction \n");
     printk("inst, par: %x %x %x \n",inst,par,cc_inst_buf[i-1]);
    }
  }
  return;
}



